@@ -1,5 +1,5 @@
 /* Compute MD5 or SHA1 checksum of files or strings
-   Copyright (C) 1995-2004 Free Software Foundation, Inc.
+   Copyright (C) 1995-2005 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,14 +13,13 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation,
-   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
 
 /* Written by Ulrich Drepper <drepper@gnu.ai.mit.edu>.  */
 
 #include <config.h>
 
 #include <getopt.h>
-#include <stdio.h>
 #include <sys/types.h>
 
 #include "system.h"
@@ -30,38 +29,17 @@
 #include "checksum.h"
 #include "getline.h"
 #include "error.h"
+#include "quote.h"
+#include "stdio--.h"
 
 /* The official name of this program (e.g., no `g' prefix).  */
-#define PROGRAM_NAME (algorithm == ALG_MD5 ? "md5sum" : "shasum")
+#define PROGRAM_NAME (algorithm == ALG_MD5 ? "md5sum" : "sha1sum")
 
 #define AUTHORS "Ulrich Drepper", "Scott Miller"
 
-/* Most systems do not distinguish between external and internal
-   text representations.  */
-/* FIXME: This begs for an autoconf test.  */
-#if O_BINARY
-# define OPENOPTS(BINARY) ((BINARY) != 0 ? TEXT1TO1 : TEXTCNVT)
-# define TEXT1TO1 "rb"
-# define TEXTCNVT "r"
-#else
-# if defined VMS
-#  define OPENOPTS(BINARY) ((BINARY) != 0 ? TEXT1TO1 : TEXTCNVT)
-#  define TEXT1TO1 "rb", "ctx=stm"
-#  define TEXTCNVT "r", "ctx=stm"
-# else
-#  if UNIX || __UNIX__ || unix || __unix__ || _POSIX_VERSION
-#   define OPENOPTS(BINARY) "r"
-#  else
-    /* The following line is intended to evoke an error.
-       Using #error is not portable enough.  */
-    "Cannot determine system type."
-#  endif
-# endif
-#endif
-
 
 #define DIGEST_TYPE_STRING(Alg) ((Alg) == ALG_MD5 ? "MD5" : "SHA1")
-#define DIGEST_STREAM(Alg) ((Alg) == ALG_MD5 ? md5_stream : sha_stream)
+#define DIGEST_STREAM(Alg) ((Alg) == ALG_MD5 ? md5_stream : sha1_stream)
 
 #define DIGEST_BITS(Alg) ((Alg) == ALG_MD5 ? 128 : 160)
 #define DIGEST_HEX_BYTES(Alg) (DIGEST_BITS (Alg) / 4)
@@ -77,8 +55,8 @@
    + 2 /* blank and binary indicator */ \
    + 1 /* minimum filename length */ )
 
-/* Nonzero if any of the files read were the standard input. */
-static int have_read_stdin;
+/* True if any of the files read were the standard input. */
+static bool have_read_stdin;
 
 /* The minimum length of a valid checksum line for the selected algorithm.  */
 static size_t min_digest_line_length;
@@ -88,11 +66,11 @@ static size_t digest_hex_bytes;
 
 /* With --check, don't generate any output.
    The exit code indicates success or failure.  */
-static int status_only = 0;
+static bool status_only = false;
 
 /* With --check, print a message to standard error warning about each
    improperly formatted checksum line.  */
-static int warn = 0;
+static bool warn = false;
 
 /* Declared and set via one of the wrapper .c files.  */
 /* int algorithm = ALG_UNSPECIFIED; */
@@ -100,14 +78,20 @@ static int warn = 0;
 /* The name this program was run with.  */
 char *program_name;
 
+/* For long options that have no equivalent short option, use a
+   non-character as a pseudo short option, starting with CHAR_MAX + 1.  */
+enum
+{
+  STATUS_OPTION = CHAR_MAX + 1
+};
+
 static const struct option long_options[] =
 {
-  { "binary", no_argument, 0, 'b' },
-  { "check", no_argument, 0, 'c' },
-  { "status", no_argument, 0, 2 },
-  { "string", required_argument, 0, 1 },
-  { "text", no_argument, 0, 't' },
-  { "warn", no_argument, 0, 'w' },
+  { "binary", no_argument, NULL, 'b' },
+  { "check", no_argument, NULL, 'c' },
+  { "status", no_argument, NULL, STATUS_OPTION },
+  { "text", no_argument, NULL, 't' },
+  { "warn", no_argument, NULL, 'w' },
   { GETOPT_HELP_OPTION_DECL },
   { GETOPT_VERSION_OPTION_DECL },
   { NULL, 0, NULL, 0 }
@@ -123,25 +107,37 @@ usage (int status)
     {
       printf (_("\
 Usage: %s [OPTION] [FILE]...\n\
-  or:  %s [OPTION] --check [FILE]\n\
 Print or check %s (%d-bit) checksums.\n\
 With no FILE, or when FILE is -, read standard input.\n\
+\n\
 "),
-	      program_name, program_name,
+	      program_name,
 	      DIGEST_TYPE_STRING (algorithm),
 	      DIGEST_BITS (algorithm));
+      if (O_BINARY)
+	fputs (_("\
+  -b, --binary            read in binary mode (default unless reading tty stdin)\n\
+"), stdout);
+      else
+	fputs (_("\
+  -b, --binary            read in binary mode\n\
+"), stdout);
       printf (_("\
-\n\
-  -b, --binary            read files in binary mode (default on DOS/Windows)\n\
-  -c, --check             check %s sums against given list\n\
-  -t, --text              read files in text mode (default)\n\
-\n\
-"),
+  -c, --check             read %s sums from the FILEs and check them\n"),
 	      DIGEST_TYPE_STRING (algorithm));
+      if (O_BINARY)
+	fputs (_("\
+  -t, --text              read in text mode (default if reading tty stdin)\n\
+"), stdout);
+      else
+	fputs (_("\
+  -t, --text              read in text mode (default)\n\
+"), stdout);
       fputs (_("\
+\n\
 The following two options are useful only when verifying checksums:\n\
       --status            don't output anything, status code shows success\n\
-  -w, --warn              warn about improperly formated checksum lines\n\
+  -w, --warn              warn about improperly formatted checksum lines\n\
 \n\
 "), stdout);
       fputs (HELP_OPTION_DESCRIPTION, stdout);
@@ -163,9 +159,9 @@ text), and name for each FILE.\n"),
 
 /* Split the checksum string S (of length S_LEN) from a BSD 'md5' or
    'sha1' command into two parts: a hexadecimal digest, and the file
-   name.  S is modified.  */
+   name.  S is modified.  Return true if successful.  */
 
-static int
+static bool
 bsd_split_3 (char *s, size_t s_len, unsigned char **hex_digest, char **file_name)
 {
   size_t i;
@@ -179,7 +175,7 @@ bsd_split_3 (char *s, size_t s_len, unsigned char **hex_digest, char **file_name
     i--;
 
   if (s[i] != ')')
-    return 1;
+    return false;
 
   s[i++] = '\0';
 
@@ -187,7 +183,7 @@ bsd_split_3 (char *s, size_t s_len, unsigned char **hex_digest, char **file_name
     i++;
 
   if (s[i] != '=')
-    return 1;
+    return false;
 
   i++;
 
@@ -195,19 +191,19 @@ bsd_split_3 (char *s, size_t s_len, unsigned char **hex_digest, char **file_name
     i++;
 
   *hex_digest = (unsigned char *) &s[i];
-  return 0;
+  return true;
 }
 
 /* Split the string S (of length S_LEN) into three parts:
    a hexadecimal digest, binary flag, and the file name.
-   S is modified.  */
+   S is modified.  Return true if successful.  */
 
-static int
+static bool
 split_3 (char *s, size_t s_len,
 	 unsigned char **hex_digest, int *binary, char **file_name)
 {
   size_t i;
-  int escaped_filename = 0;
+  bool escaped_filename = false;
   size_t algo_name_len;
 
   i = 0;
@@ -232,12 +228,12 @@ split_3 (char *s, size_t s_len,
      the first is a backslash) more characters to contain correct message digest
      information.  */
   if (s_len - i < min_digest_line_length + (s[i] == '\\'))
-    return 1;
+    return false;
 
   if (s[i] == '\\')
     {
       ++i;
-      escaped_filename = 1;
+      escaped_filename = true;
     }
   *hex_digest = (unsigned char *) &s[i];
 
@@ -246,12 +242,12 @@ split_3 (char *s, size_t s_len,
      immediately by a white space it's an error.  */
   i += digest_hex_bytes;
   if (!ISWHITE (s[i]))
-    return 1;
+    return false;
 
   s[i++] = '\0';
 
   if (s[i] != ' ' && s[i] != '*')
-    return 1;
+    return false;
   *binary = (s[i++] == '*');
 
   /* All characters between the type indicator and end of line are
@@ -273,7 +269,7 @@ split_3 (char *s, size_t s_len,
 	      if (i == s_len - 1)
 		{
 		  /* A valid line does not end with a backslash.  */
-		  return 1;
+		  return false;
 		}
 	      ++i;
 	      switch (s[i++])
@@ -286,13 +282,13 @@ split_3 (char *s, size_t s_len,
 		  break;
 		default:
 		  /* Only `\' or `n' may follow a backslash.  */
-		  return 1;
+		  return false;
 		}
 	      break;
 
 	    case '\0':
 	      /* The file name may not contain a NUL.  */
-	      return 1;
+	      return false;
 	      break;
 
 	    default:
@@ -302,55 +298,60 @@ split_3 (char *s, size_t s_len,
 	}
       *dst = '\0';
     }
-  return 0;
+  return true;
 }
 
-static int
+static bool
 hex_digits (unsigned char const *s)
 {
   while (*s)
     {
       if (!ISXDIGIT (*s))
-        return 0;
+        return false;
       ++s;
     }
-  return 1;
+  return true;
 }
 
 /* An interface to the function, DIGEST_STREAM,
-   (either md5_stream or sha_stream).
-   Operate on FILENAME (it may be "-") and put the result in *BIN_RESULT.
-   Return non-zero upon failure, zero to indicate success.  */
+   (either md5_stream or sha1_stream).
+   Operate on FILENAME (it may be "-").
 
-static int
-digest_file (const char *filename, int binary, unsigned char *bin_result,
-	   int (*digest_stream)(FILE *, void *))
+   *BINARY indicates whether the file is binary.  BINARY < 0 means it
+   depends on whether binary mode makes any difference and the file is
+   a terminal; in that case, clear *BINARY if the file was treated as
+   text because it was a terminal.
+
+   Put the checksum in *BIN_RESULT.
+   Return true if successful.  */
+
+static bool
+digest_file (const char *filename, int *binary, unsigned char *bin_result,
+	     int (*digest_stream) (FILE *, void *))
 {
   FILE *fp;
   int err;
+  bool is_stdin = STREQ (filename, "-");
 
-  if (STREQ (filename, "-"))
+  if (is_stdin)
     {
-      have_read_stdin = 1;
+      have_read_stdin = true;
       fp = stdin;
-#if O_BINARY
-      /* If we need binary reads from a pipe or redirected stdin, we need
-	 to switch it to BINARY mode here, since stdin is already open.  */
-      if (binary)
-	SET_BINARY (fileno (stdin));
-#endif
+      if (O_BINARY && *binary)
+	{
+	  if (*binary < 0)
+	    *binary = ! isatty (STDIN_FILENO);
+	  if (*binary)
+	    freopen (NULL, "rb", stdin);
+	}
     }
   else
     {
-      /* OPENOPTS is a macro.  It varies with the system.
-	 Some systems distinguish between internal and
-	 external text representations.  */
-
-      fp = fopen (filename, OPENOPTS (binary));
+      fp = fopen (filename, (O_BINARY && *binary ? "rb" : "r"));
       if (fp == NULL)
 	{
 	  error (0, errno, "%s", filename);
-	  return 1;
+	  return false;
 	}
     }
 
@@ -360,33 +361,34 @@ digest_file (const char *filename, int binary, unsigned char *bin_result,
       error (0, errno, "%s", filename);
       if (fp != stdin)
 	fclose (fp);
-      return 1;
+      return false;
     }
 
-  if (fp != stdin && fclose (fp) == EOF)
+  if (!is_stdin && fclose (fp) != 0)
     {
       error (0, errno, "%s", filename);
-      return 1;
+      return false;
     }
 
-  return 0;
+  return true;
 }
 
-static int
-digest_check (const char *checkfile_name, int (*digest_stream)(FILE *, void *))
+static bool
+digest_check (const char *checkfile_name, int (*digest_stream) (FILE *, void *))
 {
   FILE *checkfile_stream;
-  int n_properly_formated_lines = 0;
-  int n_mismatched_checksums = 0;
-  int n_open_or_read_failures = 0;
+  uintmax_t n_properly_formatted_lines = 0;
+  uintmax_t n_mismatched_checksums = 0;
+  uintmax_t n_open_or_read_failures = 0;
   unsigned char bin_buffer[MAX_DIGEST_BIN_BYTES];
-  size_t line_number;
+  uintmax_t line_number;
   char *line;
   size_t line_chars_allocated;
+  bool is_stdin = STREQ (checkfile_name, "-");
 
-  if (STREQ (checkfile_name, "-"))
+  if (is_stdin)
     {
-      have_read_stdin = 1;
+      have_read_stdin = true;
       checkfile_name = _("standard input");
       checkfile_stream = stdin;
     }
@@ -396,11 +398,10 @@ digest_check (const char *checkfile_name, int (*digest_stream)(FILE *, void *))
       if (checkfile_stream == NULL)
 	{
 	  error (0, errno, "%s", checkfile_name);
-	  return 1;
+	  return false;
 	}
     }
 
-  SET_MODE (fileno (checkfile_stream), O_TEXT);
   line_number = 0;
   line = NULL;
   line_chars_allocated = 0;
@@ -408,11 +409,13 @@ digest_check (const char *checkfile_name, int (*digest_stream)(FILE *, void *))
     {
       char *filename;
       int binary;
-      unsigned char *hex_digest;
-      int err;
-      int line_length;
+      unsigned char *hex_digest IF_LINT (= NULL);
+      ssize_t line_length;
 
       ++line_number;
+      if (line_number == 0)
+	error (EXIT_FAILURE, 0, _("%s: too many checksum lines"),
+	       checkfile_name);
 
       line_length = getline (&line, &line_chars_allocated, checkfile_stream);
       if (line_length <= 0)
@@ -426,14 +429,16 @@ digest_check (const char *checkfile_name, int (*digest_stream)(FILE *, void *))
       if (line[line_length - 1] == '\n')
 	line[--line_length] = '\0';
 
-      err = split_3 (line, line_length, &hex_digest, &binary, &filename);
-      if (err || !hex_digits (hex_digest))
+      if (! (split_3 (line, line_length, &hex_digest, &binary, &filename)
+	     && ! (is_stdin && STREQ (filename, "-"))
+	     && hex_digits (hex_digest)))
 	{
 	  if (warn)
 	    {
 	      error (0, 0,
-		     _("%s: %lu: improperly formatted %s checksum line"),
-		     checkfile_name, (unsigned long) line_number,
+		     _("%s: %" PRIuMAX
+		       ": improperly formatted %s checksum line"),
+		     checkfile_name, line_number,
 		     DIGEST_TYPE_STRING (algorithm));
 	    }
 	}
@@ -443,13 +448,13 @@ digest_check (const char *checkfile_name, int (*digest_stream)(FILE *, void *))
 					  '4', '5', '6', '7',
 					  '8', '9', 'a', 'b',
 					  'c', 'd', 'e', 'f' };
-	  int fail;
+	  bool ok;
 
-	  ++n_properly_formated_lines;
+	  ++n_properly_formatted_lines;
 
-	  fail = digest_file (filename, binary, bin_buffer, digest_stream);
+	  ok = digest_file (filename, &binary, bin_buffer, digest_stream);
 
-	  if (fail)
+	  if (!ok)
 	    {
 	      ++n_open_or_read_failures;
 	      if (!status_only)
@@ -486,22 +491,21 @@ digest_check (const char *checkfile_name, int (*digest_stream)(FILE *, void *))
     }
   while (!feof (checkfile_stream) && !ferror (checkfile_stream));
 
-  if (line)
-    free (line);
+  free (line);
 
   if (ferror (checkfile_stream))
     {
       error (0, 0, _("%s: read error"), checkfile_name);
-      return 1;
+      return false;
     }
 
-  if (checkfile_stream != stdin && fclose (checkfile_stream) == EOF)
+  if (!is_stdin && fclose (checkfile_stream) != 0)
     {
       error (0, errno, "%s", checkfile_name);
-      return 1;
+      return false;
     }
 
-  if (n_properly_formated_lines == 0)
+  if (n_properly_formatted_lines == 0)
     {
       /* Warn if no tests are found.  */
       error (0, 0, _("%s: no properly formatted %s checksum lines found"),
@@ -511,52 +515,43 @@ digest_check (const char *checkfile_name, int (*digest_stream)(FILE *, void *))
     {
       if (!status_only)
 	{
-	  int n_computed_checkums = (n_properly_formated_lines
-				     - n_open_or_read_failures);
+	  if (n_open_or_read_failures != 0)
+	    error (0, 0,
+		   ngettext ("WARNING: %" PRIuMAX " of %" PRIuMAX
+			     " listed file could not be read",
+			     "WARNING: %" PRIuMAX " of %" PRIuMAX
+			     " listed files could not be read",
+			     n_properly_formatted_lines),
+		   n_open_or_read_failures, n_properly_formatted_lines);
 
-	  if (n_open_or_read_failures > 0)
+	  if (n_mismatched_checksums != 0)
 	    {
+	      uintmax_t n_computed_checksums =
+		(n_properly_formatted_lines - n_open_or_read_failures);
 	      error (0, 0,
-		     _("WARNING: %d of %d listed %s could not be read"),
-		     n_open_or_read_failures, n_properly_formated_lines,
-		     (n_properly_formated_lines == 1
-		      ? _("file") : _("files")));
-	    }
-
-	  if (n_mismatched_checksums > 0)
-	    {
-	      error (0, 0,
-		   _("WARNING: %d of %d computed %s did NOT match"),
-		     n_mismatched_checksums, n_computed_checkums,
-		     (n_computed_checkums == 1
-		      ? _("checksum") : _("checksums")));
+		     ngettext ("WARNING: %" PRIuMAX " of %" PRIuMAX
+			       " computed checksum did NOT match",
+			       "WARNING: %" PRIuMAX " of %" PRIuMAX
+			       " computed checksums did NOT match",
+			       n_computed_checksums),
+		     n_mismatched_checksums, n_computed_checksums);
 	    }
 	}
     }
 
-  return ((n_properly_formated_lines > 0 && n_mismatched_checksums == 0
-	   && n_open_or_read_failures == 0) ? 0 : 1);
+  return (n_properly_formatted_lines != 0
+	  && n_mismatched_checksums == 0
+	  && n_open_or_read_failures == 0);
 }
 
 int
 main (int argc, char **argv)
 {
   unsigned char bin_buffer[MAX_DIGEST_BIN_BYTES];
-  int do_check = 0;
+  bool do_check = false;
   int opt;
-  char **string = NULL;
-  size_t n_strings = 0;
-  int err = 0;
-  int file_type_specified = 0;
-
-#if O_BINARY
-  /* Binary is default on MSDOS, so the actual file contents
-     are used in computation.  */
-  int binary = 1;
-#else
-  /* Text is default of the Plumb/Lankester format.  */
-  int binary = 0;
-#endif
+  bool ok = true;
+  int binary = -1;
 
   /* Setting values of global variables.  */
   initialize_main (&argc, &argv);
@@ -570,36 +565,22 @@ main (int argc, char **argv)
   while ((opt = getopt_long (argc, argv, "bctw", long_options, NULL)) != -1)
     switch (opt)
       {
-      case 0:			/* long option */
-	break;
-      case 1: /* --string */
-	{
-	  if (string == NULL)
-	    string = xmalloc ((argc - 1) * sizeof (char *));
-
-	  if (optarg == NULL)
-	    optarg = "";
-	  string[n_strings++] = optarg;
-	}
-	break;
       case 'b':
-	file_type_specified = 1;
 	binary = 1;
 	break;
       case 'c':
-	do_check = 1;
+	do_check = true;
 	break;
-      case 2:
-	status_only = 1;
-	warn = 0;
+      case STATUS_OPTION:
+	status_only = true;
+	warn = false;
 	break;
       case 't':
-	file_type_specified = 1;
 	binary = 0;
 	break;
       case 'w':
-	status_only = 0;
-	warn = 1;
+	status_only = false;
+	warn = true;
 	break;
       case_GETOPT_HELP_CHAR;
       case_GETOPT_VERSION_CHAR (PROGRAM_NAME, AUTHORS);
@@ -610,83 +591,44 @@ main (int argc, char **argv)
   min_digest_line_length = MIN_DIGEST_LINE_LENGTH (algorithm);
   digest_hex_bytes = DIGEST_HEX_BYTES (algorithm);
 
-  if (file_type_specified && do_check)
+  if (0 <= binary && do_check)
     {
-      error (0, 0, _("the --binary and --text options are meaningless when \
-verifying checksums"));
+      error (0, 0, _("the --binary and --text options are meaningless when "
+		     "verifying checksums"));
       usage (EXIT_FAILURE);
     }
 
-  if (n_strings > 0 && do_check)
-    {
-      error (0, 0,
-	     _("the --string and --check options are mutually exclusive"));
-      usage (EXIT_FAILURE);
-    }
-
-  if (status_only && !do_check)
+  if (status_only & !do_check)
     {
       error (0, 0,
        _("the --status option is meaningful only when verifying checksums"));
       usage (EXIT_FAILURE);
     }
 
-  if (warn && !do_check)
+  if (warn & !do_check)
     {
       error (0, 0,
        _("the --warn option is meaningful only when verifying checksums"));
       usage (EXIT_FAILURE);
     }
 
-  if (n_strings > 0)
-    {
-      size_t i;
+  if (optind == argc)
+    argv[argc++] = "-";
 
-      if (optind < argc)
+  for (; optind < argc; ++optind)
+    {
+      char *file = argv[optind];
+
+      if (do_check)
+	ok &= digest_check (file, DIGEST_STREAM (algorithm));
+      else
 	{
-	  error (0, 0, _("no files may be specified when using --string"));
-	  usage (EXIT_FAILURE);
-	}
-      for (i = 0; i < n_strings; ++i)
-	{
-	  size_t cnt;
-	  if (algorithm == ALG_MD5)
-	    md5_buffer (string[i], strlen (string[i]), bin_buffer);
+	  int file_is_binary = binary;
+
+	  if (! digest_file (file, &file_is_binary, bin_buffer,
+			     DIGEST_STREAM (algorithm)))
+	    ok = false;
 	  else
-	    sha_buffer (string[i], strlen (string[i]), bin_buffer);
-
-	  for (cnt = 0; cnt < (digest_hex_bytes / 2); ++cnt)
-	    printf ("%02x", bin_buffer[cnt]);
-
-	  printf ("  \"%s\"\n", string[i]);
-	}
-    }
-  else if (do_check)
-    {
-      if (optind + 1 < argc)
-	{
-	  error (0, 0,
-		 _("only one argument may be specified when using --check"));
-	  usage (EXIT_FAILURE);
-	}
-
-      err = digest_check ((optind == argc) ? "-" : argv[optind],
-			  DIGEST_STREAM (algorithm));
-    }
-  else
-    {
-      if (optind == argc)
-	argv[argc++] = "-";
-
-      for (; optind < argc; ++optind)
-	{
-	  int fail;
-	  char *file = argv[optind];
-
-	  fail = digest_file (file, binary, bin_buffer,
-			      DIGEST_STREAM (algorithm));
-	  err |= fail;
-	  if (!fail)
 	    {
 	      size_t i;
 
@@ -699,7 +641,7 @@ verifying checksums"));
 		printf ("%02x", bin_buffer[i]);
 
 	      putchar (' ');
-	      if (binary)
+	      if (file_is_binary)
 		putchar ('*');
 	      else
 		putchar (' ');
@@ -731,5 +673,5 @@ verifying checksums"));
   if (have_read_stdin && fclose (stdin) == EOF)
     error (EXIT_FAILURE, errno, _("standard input"));
 
-  exit (err == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
+  exit (ok ? EXIT_SUCCESS : EXIT_FAILURE);
 }
